from osgeo import gdal
import numpy as np
from pathlib import Path
from typing import Union, Optional
from scipy.interpolate import Rbf
import fluxpark as flp


def interpolate_rain(
    x,
    y,
    point_values,
    dst_epsg,
    bounds,
    cellsize,
    radius=50000.0,
    max_points=3,
    coarse_cellsize=2500.0,
    intermediate_path: Optional[Union[str, Path]] = None,
):
    """
    Interpolate point rainfall values to a raster grid using gdal.Grid,
    then resample to coarser grid and smooth back to target resolution.

    Parameters
    ----------
    x, y : array-like
        Coordinates of the points.
    values : array-like
        Rainfall values.
    dst_epsg : int
        EPSG code of output projection.
    bounds : tuple
        (x_min, x_max, y_min, y_max) in output CRS.
    cellsize : float
        Final output resolution (meters).
    radius : float
        Search radius for IDW interpolation.
    max_points : int
        Max number of neighbors in interpolation.
    coarse_cellsize : float
        Intermediate resolution (default: 2500m) to speed up interpolation.
    intermediate_path : str or Path, optional
        Full filename (including directory) where intermediate outputs
        will be saved (e.g. a GeoPackage or VRT).  If the directory
        doesn’t exist, it will be created. If None (the default),
        no intermediate files are written.

    Returns
    -------
    np.ndarray
        Interpolated raster as NumPy array (np.nan = no data).
    """
    x_min, x_max, y_min, y_max = bounds

    # 1. create a point data set.
    ds_point = flp.io.write_point_layer(x, y, point_values, dst_epsg, intermediate_path)

    # 2. Interpolate to coarse grid using IDW
    ncols = int((x_max - x_min) / coarse_cellsize) + 1
    nrows = int((y_max - y_min) / coarse_cellsize) + 1
    x_max_snap = x_min + ncols * coarse_cellsize
    y_min_snap = y_max - nrows * coarse_cellsize

    ds_initial = gdal.Grid(
        "",
        ds_point,
        format="MEM",
        outputBounds=[x_min, y_max, x_max_snap, y_min_snap],
        width=ncols,
        height=nrows,
        outputType=gdal.GDT_Float32,
        algorithm=(
            f"invdistnn:power=2.0:smoothing=0.0:"
            f"radius={radius}:max_points={max_points}:"
            f"min_points=0:nodata=-9999"
        ),
        zfield="z",
    )

    # 3. Warp to coarser grid with averaging
    x_max_coarse = (
        x_min + (int((x_max - x_min) / coarse_cellsize) + 1) * coarse_cellsize
    )
    y_min_coarse = (
        y_max - (int((y_max - y_min) / coarse_cellsize) + 1) * coarse_cellsize
    )

    ds_coarse = gdal.Warp(
        "",
        ds_initial,
        dstSRS=f"EPSG:{dst_epsg}",
        format="VRT",
        resampleAlg=gdal.GRA_Average,
        xRes=coarse_cellsize,
        yRes=-coarse_cellsize,
        outputBounds=[x_min, y_min_coarse, x_max_coarse, y_max],
        outputBoundsSRS=f"EPSG:{dst_epsg}",
        dstNodata=-9999,
    )

    # 4. Warp back to fine resolution with smoothing
    ds_fine = gdal.Warp(
        "",
        ds_coarse,
        dstSRS=f"EPSG:{dst_epsg}",
        format="VRT",
        resampleAlg=gdal.GRA_CubicSpline,
        xRes=cellsize,
        yRes=-cellsize,
        outputBounds=[x_min, y_min, x_max, y_max],
        outputBoundsSRS=f"EPSG:{dst_epsg}",
        dstNodata=-9999,
    )

    # 5. Read array and clean up
    rain = ds_fine.ReadAsArray().astype(np.float32)
    rain[rain == -9999] = np.nan
    rain[rain < 0.1] = 0.0

    # Clean memory
    ds_point = None
    ds_initial = None
    ds_coarse = None
    ds_fine = None

    return rain


def interpolate_makkink(
    x: np.ndarray,
    y: np.ndarray,
    point_values: np.ndarray,
    dst_epsg: int,
    bounds: tuple[float, float, float, float],
    cellsize: float,
    coarse_cellsize: float = 2500.0,
    intermediate_path: Optional[Union[str, Path]] = None,
) -> np.ndarray:
    """
    Interpolate Makkink reference ET to a raster using
    a thin-plate spline RBF on a coarse grid + GDAL Warp to
    resample back to full resolution.

    Parameters
    ----------
    x, y : array-like
        Coordinates of observation points (in output CRS).
    mak_values : array-like
        Makkink ET values at those points.
    dst_epsg : int
        EPSG code of output projection.
    bounds : (x_min, x_max, y_min, y_max)
        Target extent.
    cellsize : float
        Final output resolution (meters).
    coarse_cellsize : float, optional
        Resolution of the intermediate grid (default: 2500 m).
    intermediate_path : str or Path, optional
        If provided, writes your **input points** (with field "mak")
        to the given filename (GeoPackage, Shapefile, etc.).  No
        intermediate grids are written. Parent directory is created
        if needed. If None, no files are written.
    Returns
    -------
    np.ndarray
        Interpolated raster (dtype float32).
    """
    x_min, x_max, y_min, y_max = bounds

    # 1. Build the RBF interpolator
    rbf = Rbf(x, y, point_values, function="thin_plate", smooth=0)

    # 2. Define & populate coarse grid
    nx = int(np.ceil((x_max - x_min) / coarse_cellsize)) + 1
    ny = int(np.ceil((y_max - y_min) / coarse_cellsize)) + 1
    x_max_snap = x_min + (nx - 1) * coarse_cellsize
    y_min_snap = y_max - (ny - 1) * coarse_cellsize

    xs = np.linspace(x_min, x_max_snap, nx)
    ys = np.linspace(y_max, y_min_snap, ny)  # descending
    xi, yi = np.meshgrid(xs, ys)
    mak_course = rbf(xi, yi).astype(np.float32)

    # 3. Create an in-memory GeoTIFF from that coarse array
    mak_course_ds = flp.io.write_geotiff(
        "", "", mak_course, x_min, y_max, coarse_cellsize, dst_epsg
    )

    # 4. Warp back to full resolution with cubic-spline resampling
    ds_fine = gdal.Warp(
        "",
        mak_course_ds,
        dstSRS=f"EPSG:{dst_epsg}",
        format="VRT",
        resampleAlg=gdal.GRA_CubicSpline,
        xRes=cellsize,
        yRes=-cellsize,
        outputBounds=[x_min, y_min, x_max, y_max],
        outputBoundsSRS=f"EPSG:{dst_epsg}",
        dstNodata=-9999,
    )

    # 5. Read and post-process
    mak = ds_fine.ReadAsArray().astype(np.float32)
    mak[mak < 0.0] = 0.05
    mn, mx = point_values.min(), point_values.max()
    mak = np.clip(mak, mn * 0.99, mx * 1.01)

    # 6. create a point data set for analysis.
    if intermediate_path:
        flp.io.write_point_layer(x, y, point_values, dst_epsg, intermediate_path)

    # 7. Clean up
    mak_course_ds = None
    ds_fine = None

    return mak



"""
Computation of root water uptake.

This module provides functions to estimate soil moisture deficit and actual
transpiration due to root water uptake.

Functions
---------
unsat_reservoirmodel(rain, etp, smd_old, rc, pwp)
    Calculates actual transpiration.
"""

import numpy as np


def unsat_reservoirmodel(rain, etp, smd_old, soilm_scp, soilm_pwp):
    """
    Unsaturated zone model to estimate actual transpiration.

    Simulates actual evapotranspiration (eta), potential soil moisture deficit
    (smdp), updated soil moisture deficit (smda), and drainage using a simple
    unsaturated zone reservoir model. Based on: Ireson, A.M., Butler, A.P.,
    2013. A critical assessment of simple recharge models. Hydrol. Earth Syst.
    Sci., 17(6): 2083–2096.

    Parameters
    ----------
    rain : ndarray
        Precipitation or throughfall input [L per day]. 1D or 2D array.
    etp : ndarray
        Potential evapotranspiration from the unsaturated zone [L per day].
    smd_old : ndarray
        Previous soil moisture deficit [L].
    soilm_scp : ndarray
        Stomatal closure point [L]. The root zone soil moisture deficit at
        which ET starts to reduce.
    soilm_pwp : ndarray
        Permanent wilting point [L]. The root zone soil moisture deficit at
        which ET ceases entirely.

    Returns
    -------
    eta : ndarray
        Actual evapotranspiration [L per day], same shape as input.
    smdp : ndarray
        Potential soil moisture deficit [L].
    smda : ndarray
        Updated soil moisture deficit [L].
    drainage : ndarray
        Drainage out of the unsaturated zone [L per day].
    """
    # Input validation
    if not (
        rain.shape == etp.shape
        and etp.shape == smd_old.shape
        and smd_old.shape == soilm_scp.shape
        and soilm_scp.shape == soilm_pwp.shape
    ):
        raise ValueError("All input arrays must have the same shape.")

    # Get input shape
    shape = rain.shape

    # Initialize output arrays
    eta = np.zeros(shape, dtype=np.float32)
    drainage = np.zeros(shape, dtype=np.float32)
    smda = np.zeros(shape, dtype=np.float32)

    # Compute potential soil moisture deficit
    smdp = smd_old - rain + etp

    # Evaporation conditions
    cond1 = smdp <= soilm_scp
    cond2 = (smdp > soilm_scp) & (smdp < soilm_pwp)
    # cond3 = ~(cond1 | cond2)

    # Calculate actual evapotranspiration
    eta[cond1] = etp[cond1]
    eta[cond2] = etp[cond2] * (
        (soilm_pwp[cond2] - smdp[cond2]) / (soilm_pwp[cond2] - soilm_scp[cond2])
    )

    # eta[cond3] remains zero

    # Drainage and updated SMD
    cond_drain = smdp < 0
    cond_no_drain = ~cond_drain

    drainage[cond_drain] = -smdp[cond_drain]
    # drainage[cond_no_drain] remains zero

    smda[cond_drain] = 0.0
    smda[cond_no_drain] = (
        smd_old[cond_no_drain] - rain[cond_no_drain] + eta[cond_no_drain]
    )

    return eta, smdp, smda, drainage



"""
Computation of interception evaporation.

This module provides functions for estimating interception evaporation based on
staandard meteorological variables and surface/plant properties.

Functions
---------
interception_voortman(etp_wet, rain, max_int, cover, int_store_old)
    Calculates actual interception evaporation and throughfall based on
    Voortman (2015).
"""

import numpy as np


def interception_voortman(etp_wet, rain, max_int, cover, int_store_old):
    """
    Actual interception evaporation and throughfall based on Voortman (2015).

    Calculates actual interception evaporation, updated interception store,
    throughfall and fraction of the timestep with evaporation.

    Parameters
    ----------
    etp_wet : ndarray
        Potential interception evaporation assuming 100% vegetation cover
        (e.g. in mm or m).
    rain : ndarray
        Precipitation input (same unit as etp_wet).
    max_int : ndarray or float
        Maximum interception storage capacity for 100% cover (same unit).
    cover : ndarray or float
        Vegetation cover fraction [m²/m²].
    int_store_old : ndarray
        Interception store at the start of the timestep.

    Returns
    -------
    int_evap : ndarray
        Actual interception evaporation.
    int_store : ndarray
        Updated interception storage at the end of the timestep.
    throughfall : ndarray
        Precipitation not intercepted by vegetation.
    int_timefrac : ndarray
        Fraction of the timestep where interception evaporation occurs.

    Notes
    -----
    All input and output arrays are assumed to be 2D arrays with shape [x, y].
    """
    # Intercepted rain and capacity (adjusted for cover)
    rain_int = rain * cover
    int_cap = max_int * cover
    etp_int = etp_wet * cover

    # Addition to interception storage
    capacity = int_cap - int_store_old
    int_add = np.minimum(rain_int, capacity)
    int_store = int_store_old + int_add

    # Interception evaporation logic (cannot exceed int_store)
    int_evap = np.minimum(etp_int, int_store)

    # Update interception store after evaporation
    int_store = int_store - int_evap

    # Throughfall (rain not intercepted)
    throughfall = rain - int_add

    # Calculate fraction of timestep with interception evaporation
    with np.errstate(divide="ignore", invalid="ignore"):
        int_timefrac = np.divide(
            int_evap, etp_int, out=np.zeros_like(etp_int), where=etp_int != 0
        )

    return int_evap, int_store, throughfall, int_timefrac



from pathlib import Path
import math
import numpy as np
from numpy.typing import NDArray
import pandas as pd
from typing import Optional, Tuple, Dict, Any, List, Union, cast
import fluxpark as flp
import logging
import os


def check_output_files(
    output_files: str, conv_output_df: pd.DataFrame, mod_flags: dict, store_states: bool
):
    """
    Determine which parameters should be written, calculated, and aggregated.

    Parameters
    ----------
    output_files : str or list of int
        Desired output set. Can be:
        - "all" to include all available parameters.
        - "flagship" to include key parameters.
        - "selection" for your own preference.
        - A list of row indices from the conversion table.
    conv_output_df : pandas.DataFrame
        Conversion table with parameter names as index and
        module flags (e.g., 'mod_crop_prod') as columns.
    mod_flags : dict
        Dictionary mapping module names to bool flags,
        e.g. {'mod_dynamic_grass': True, 'mod_seen': False}
    store_states: bool
        If true, calculation parameters are added to the output list.

    Returns
    -------
    output_par_list : list of str
        List of Nexus parameters to write as output.
    calc_par_list : list of str
        Parameters required for internal calculations.
    cum_par_list : list of str
        Subset of output_par_list requiring aggregation.
    """
    # Core model selection
    conv_output_df_sel = conv_output_df[conv_output_df["mod"] == "core"]
    calc_par_list = conv_output_df[conv_output_df["calc_core"] == 1].index.tolist()

    # Extend based on active modules
    for mod_key, enabled in mod_flags.items():
        if not enabled:
            continue
        mod_col = mod_key
        calc_col = f"calc_{mod_key}"
        if mod_col in conv_output_df.columns:
            conv_output_df_sel = pd.concat(
                [conv_output_df_sel, conv_output_df[conv_output_df[mod_col] == 1]]
            )
        if calc_col in conv_output_df.columns:
            calc_par_list += conv_output_df[
                conv_output_df[calc_col] == 1
            ].index.tolist()

    # Determine output parameters
    if output_files == "all":
        output_par_list = conv_output_df_sel.index.tolist()
    elif output_files == "flagship":
        output_par_list = conv_output_df_sel[
            conv_output_df_sel["flagship"] == 1
        ].index.tolist()
    elif output_files == "selection":
        output_par_list = conv_output_df_sel[
            conv_output_df_sel["selection"] == 1
        ].index.tolist()
    elif isinstance(output_files, str):
        raise ValueError(
            f"output_files can be 'all', 'flagship' or 'selection'. Or it can be a list"
            f" with parameters names or id's. Your value '{output_files}' does not meet"
            " the requirements."
        )
    else:
        # It should be a list and we Convert integer IDs to names or keep the names
        converted = []
        index = conv_output_df.index
        for val in output_files:
            if isinstance(val, int):
                if 0 <= val < len(index):
                    converted.append(index[val])
                else:
                    raise IndexError(f"Index {val} is out of range for output files.")
            elif isinstance(val, str):
                if val not in index:
                    raise ValueError(
                        f"Parameter name '{val}' not found in conv_output_df index."
                    )
                converted.append(val)
            else:
                raise TypeError(f"Output file spec '{val}' must be int or str.")

        output_par_list = [
            name for name in converted if name in conv_output_df_sel.index
        ]

        if len(output_par_list) != len(converted):
            logging.warning(
                "Some output_files do not match the active modules and were ignored."
            )

    # Handle interdependencies
    if (
        "trans_def_cum_past10d_mm" in output_par_list
        and "trans_def_pot_mm_d" not in output_par_list
    ):
        output_par_list.append("trans_def_pot_mm_d")
    if (
        "drought_stress_index_pct" in output_par_list
        and "trans_rel_pct" not in output_par_list
    ):
        output_par_list.append("trans_rel_pct")

    # Cumulative parameters
    cum_par_list_all = conv_output_df_sel[
        conv_output_df_sel["cumulative"] == 1
    ].index.tolist()
    cum_par_list = [x for x in cum_par_list_all if x in output_par_list]

    # Force calc parameters into output list
    if store_states:
        missing = [x for x in calc_par_list if x not in output_par_list]
        output_par_list += missing

    return output_par_list, calc_par_list, cum_par_list


def detect_dynamic_landuse_and_years(
    landuse_filename, root_soilm_scp_filename, root_soilm_pwp_filename, indir_rasters
):
    """
    Determine if input maps are dynamic and list available years.

    Parameters
    ----------
    landuse_filename : str
        Filename pattern for land use maps, may include '{year}'.
    root_soilm_scp_filename : str
        Filename pattern for soil moisture SCP maps.
    root_soilm_pwp_filename : str
        Filename pattern for soil moisture PWP maps.
    indir_rasters : Path or str
        Directory containing input raster files.

    Returns
    -------
    dynamic : bool
        True if all patterns include '{year}', False otherwise.
    input_raster_years : ndarray of str
        Sorted unique list of years for which 'luse_ids.tif' files
        exist.
    """
    is_luse = flp.utils.has_placeholders(landuse_filename)
    is_scp = flp.utils.has_placeholders(root_soilm_scp_filename)
    is_pwp = flp.utils.has_placeholders(root_soilm_pwp_filename)

    if is_luse and is_scp and is_pwp:
        logging.info(
            "Run default dynamic landuse simulations that could "
            "change yearly if available"
        )
        dynamic = True
    elif is_luse or is_scp or is_pwp:
        raise RuntimeError(
            f"Not all input maps have '{{year}}' placeholder: "
            f"{landuse_filename}, {root_soilm_scp_filename}, "
            f"{root_soilm_pwp_filename}. Either all are static or all "
            "are dynamic"
        )
    else:
        logging.info("Run with static landuse map")
        dynamic = False

    files = os.listdir(indir_rasters)
    mask = ["luse_ids.tif" in f for f in files]
    years = np.array([f.split("_")[0] for f in np.array(files)[mask]])
    input_raster_years = np.sort(np.unique(years))

    return dynamic, input_raster_years


def parse_dates(start: str, end: str) -> pd.DatetimeIndex:
    """
    Parse start and end dates into a daily pandas DateTimeIndex.

    Parameters
    ----------
    start : str
        Start date string (e.g. '2025-06-01').
    end : str
        End date string.

    Returns
    -------
    pd.DatetimeIndex
        Sequence of dates from start to end inclusive.
    """
    s = pd.to_datetime(start, dayfirst=True, errors="raise")
    e = pd.to_datetime(end, dayfirst=True, errors="raise")
    return pd.date_range(s, e, freq="D")


def resolve_dirs(
    outdir: Union[str, Path],
    indir: Union[str, Path],
    indir_rasters: Optional[Union[str, Path]] = None,
    indir_masks: Optional[Union[str, Path]] = None,
    intermediate_dir: Optional[Union[str, Path]] = None,
) -> Tuple[Path, Path, Path, Path, Optional[Path]]:
    """
    Normalize output/input directories and derive rasters/masks subdirs.

    Parameters
    ----------
    outdir
        Base output directory.
    indir
        Base input directory.
    indir_rasters
        Optional override for the 'rasters' subdirectory.
    indir_masks
        Optional override for the 'masks' subdirectory.
    intermediate_dir
        Optional the dir to store intermediate output

    Returns
    -------
    out_p, in_p, rasters_p, masks_p
        Paths for output, input, rasters, and masks.
    """
    out_p = Path(outdir)
    in_p = Path(indir)
    rasters_p = Path(indir_rasters) if indir_rasters else in_p / "rasters"
    masks_p = Path(indir_masks) if indir_masks else in_p / "masks"
    if intermediate_dir:
        # Normalize to a Path
        intermediate_dir = Path(intermediate_dir)
    else:
        # be excplicit prevents typing problems
        intermediate_dir = None

    return out_p, in_p, rasters_p, masks_p, intermediate_dir


def compute_grid_params(
    x_min: float,
    x_max: float,
    y_min: float,
    y_max: float,
    cellsize: float,
    epsg_code: int,
    indir_masks: Path,
    mask: Optional[str] = None,
) -> Dict[str, Any]:
    """
    Compute grid dimensions and package geospatial settings.

    Parameters
    ----------
    x_min, x_max, y_min, y_max
        Bounding box in target CRS.
    cellsize
        Grid resolution (assumed square).
    epsg_code
        Output projection EPSG.
    indir_masks
        Base masks directory.
    mask
        Optional mask filename.

    Returns
    -------
    grid_params
        {
          "dst_epsg": int,
          "bounds": (x_min, x_max, y_min, y_max),
          "cellsize": float,
          "cutline_path": Optional[Path],
          "ncols": int,
          "nrows": int
        }
    """
    ncols = math.ceil((x_max - x_min) / cellsize)
    nrows = math.ceil((y_max - y_min) / cellsize)
    cutline = Path(indir_masks) / mask if mask else None

    return {
        "dst_epsg": epsg_code,
        "bounds": (x_min, x_max, y_min, y_max),
        "cellsize": cellsize,
        "cutline_path": cutline,
        "ncols": ncols,
        "nrows": nrows,
    }


def load_evap_params(indir: Path, evap_param_table: str) -> dict[str, np.ndarray]:
    """
    Load evaporation parameters per land use and day-of-year.

    Parameters
    ----------
    indir : Path
        Directory containing the Excel file.
    evap_param_table : str
        Filename of the evap params workbook.

    Returns
    -------
    params : dict
        Keys are column names, values are NumPy arrays.
    """
    sheets = pd.read_excel(  # type: ignore[call]
        indir / evap_param_table,
        sheet_name=None,
        skiprows=range(0, 12),
        usecols="A:I",
        na_values=str(-9999),
    )
    df = pd.concat(sheets, ignore_index=True)
    data = df.to_dict("list")
    raw = {k: np.array(v) for k, v in data.items()}
    return cast(dict[str, np.ndarray], raw)


def load_luse_evap_conv(
    indir: Path,
) -> tuple[NDArray[np.int_], NDArray[np.int_], NDArray[np.str_]]:
    """
    Read landuse→evap ID conversion table.

    Parameters
    ----------
    indir : Path
        Directory containing 'conv_luse_evap_ids.csv'.

    Returns
    -------
    luse_ids, evap_ids, labels : arrays
    """
    df = pd.read_csv(
        indir / "conv_luse_evap_ids.csv",
        dtype={
            "luse_id": np.int64,
            "evap_id": np.int64,
            "label": str,
        },
    )
    return df["luse_id"].to_numpy(), df["evap_id"].to_numpy(), df["label"].to_numpy()


def load_conv_output(
    indir: Path,
    output_mapping: str,
) -> tuple[pd.DataFrame, dict[str, str], dict[str, str]]:
    """
    Read conversion of model vars ↔ output filenames.

    Parameters
    ----------
    indir : Path
        Directory containing 'fluxpark_output_mapping.csv'.

    Returns
    -------
    conv_output, conv_var : dict
        param→var and var→param maps.
    """
    df = pd.read_csv(
        indir / output_mapping,
        skiprows=list(np.arange(0, 10, 1)),
        index_col="parameter",
    )
    out_map = dict(zip(df.index, df["variable"]))
    var_map = dict(zip(df["variable"], df.index))
    return df, out_map, var_map


def prepare_output_and_rerun_lists(
    mods: Dict[str, bool],
    output_files: Any,
    conv_output_table: "pd.DataFrame",
    conv_output: Dict[str, str],
    store_states: bool,
) -> Tuple[
    List[str],  # out_par_list
    List[str],  # calc_par_list
    List[str],  # cum_par_list
    List[str],  # out_var_list
    List[str],  # rerun_par_list
    List[str],  # rerun_var_list
]:
    """
    Determine which parameters to output and which to carry forward.

    Parameters
    ----------
    mods
        All mod flags (e.g. {'mod_...': True, …}).
    output_files
        Specification of desired outputs ('all', 'flagship', list, etc.).
    conv_output_table
        DataFrame mapping 'parameter' → output flags & module settings.
    conv_output
        Dict mapping parameter names → Python variable names.

    Returns
    -------
    out_par_list
        List of requested output parameters.
    calc_par_list
        Parameters needed for rerun initialization.
    cum_par_list
        Cumulative parameters (also must be output).
    out_var_list
        Python variable names corresponding to out_par_list.
    rerun_par_list
        Combined list of calc_par_list and cum_par_list (unique).
    rerun_var_list
        Python variable names corresponding to rerun_par_list.
    """
    # get the three lists from the conversion table
    out_par_list, calc_par_list, cum_par_list = check_output_files(
        output_files, conv_output_table, mods, store_states
    )

    # convert output parameters to variable names
    out_var_list = [conv_output[par] for par in out_par_list]

    # rerun parameters = calculation + cumulative
    rerun_par_list = list(set(calc_par_list) | set(cum_par_list))
    rerun_var_list = [conv_output[par] for par in rerun_par_list]

    return (
        out_par_list,
        calc_par_list,
        cum_par_list,
        out_var_list,
        rerun_par_list,
        rerun_var_list,
    )


def init_old(
    rerun_var_list: list[str], nrows: int, ncols: int
) -> dict[str, np.ndarray]:
    """
    Initialize the 'old' state dict for rerun variables.

    Parameters
    ----------
    rerun_var_list : list of str
        Python var names needing initial arrays.
    nrows, ncols : int
        Grid dimensions.

    Returns
    -------
    old : dict
        Each key→array of zeros or 0.68 for 'soil_cov'.
    """
    old = {}
    shape = (nrows, ncols)
    for key in rerun_var_list:
        if key == "soil_cov":
            old[key] = np.full(shape, 0.68, dtype="float32")
        else:
            old[key] = np.zeros(shape, dtype="float32")
    return old


def read_static_maps(
    indir_rasters: Path,
    grid_params: dict,
    mods: dict[str, bool],
    imperv_filename: str,
    decid_filename: str,
    conif_filename: str,
) -> Tuple[np.ndarray, Optional[np.ndarray], Optional[np.ndarray]]:
    """
    Load static rasters: imperviousness and (optionally) forest cover maps.

    Parameters
    ----------
    indir_rasters
        Directory where raster files live.
    grid_params
        Passed as **kwargs to GeoTiffReader.read_and_reproject.
    mods
        Module flags, e.g. {'mod_vegcover': True, …}.
    imperv_filename
        The impervious density file name
    conif_filename
        The coniferous forest soil cover file name
    decid_filename
        The decideous forest soil cover file name

    Returns
    -------
    imperv
        Impervious fraction (0 to 1).
    soil_cov_decid
        Deciduous forest cover fraction or None.
    soil_cov_conif
        Coniferous forest cover fraction or None.
    """

    # Imperviousness (0→1)
    reader = flp.io.GeoTiffReader(indir_rasters / imperv_filename, nodata_value=0)
    imperv = reader.read_and_reproject(**grid_params).astype(np.float32) / 100.0

    soil_cov_decid = None
    soil_cov_conif = None
    if mods.get("mod_vegcover", False):
        # Deciduous cover
        reader = flp.io.GeoTiffReader(indir_rasters / decid_filename, nodata_value=0)
        soil_cov_decid = reader.read_and_reproject(**grid_params).astype(np.float32)
        soil_cov_decid[soil_cov_decid == 0] = np.nan

        # Coniferous cover
        reader = flp.io.GeoTiffReader(indir_rasters / conif_filename, nodata_value=0)
        soil_cov_conif = reader.read_and_reproject(**grid_params).astype(np.float32)
        soil_cov_conif[soil_cov_conif == 0] = np.nan

    return imperv, soil_cov_decid, soil_cov_conif



from .core_initialization import (
    check_output_files,
    detect_dynamic_landuse_and_years,
    parse_dates,
    resolve_dirs,
    compute_grid_params,
    load_evap_params,
    load_luse_evap_conv,
    load_conv_output,
    prepare_output_and_rerun_lists,
    init_old,
    read_static_maps,
)

__all__ = [
    "check_output_files",
    "detect_dynamic_landuse_and_years",
    "parse_dates",
    "resolve_dirs",
    "compute_grid_params",
    "load_evap_params",
    "load_luse_evap_conv",
    "load_conv_output",
    "prepare_output_and_rerun_lists",
    "init_old",
    "read_static_maps",
]



"""
Computation of reference crop evapotranspiration (ETref).

This module provides functions for estimating daily evapotranspiration based on
staandard meteorological variables. The formulas are suited for temperate
climates like those in Western Europe.

Functions
---------
makkink(tair, rs_in)
    Calculates reference evapotranspiration using the Makkink equation.
"""

import numpy as np


def makkink(tair, rs_in):
    """
    Calculate Makkink reference evapotranspiration.

    Estimates the daily reference evapotranspiration based on air temperature
    and incoming solar radiation, using the Makkink equation. Commonly used in
    hydrological and agricultural models in temperate regions.

    Parameters
    ----------
    tair : float or np.ndarray
        Daily mean air temperature in degrees Celsius (°C).
    rs_in : float or np.ndarray
        Daily incoming global solar radiation in joules per square centimeter
        (J/cm²).

    Returns
    -------
    et_makkink : float or np.ndarray
        Reference evapotranspiration in millimeters per day (mm/day).

    where Δ is the slope of the saturation vapor pressure curve,
    γ is the psychrometric constant, λ is the latent heat of vaporization,
    and ρw is the density of water.
    """
    # Saturation vapour pressure (hPa)
    es = 6.107 * 10 ** (7.5 * tair / (237.3 + tair))
    # Slope of saturation vapour pressure curve (hPa/°C)
    d = 7.5 * 237.3 / ((237.3 + tair) ** 2) * np.log(10) * es
    # Psychrometric constant (hPa/°C)
    gamma = 0.646 + 0.0006 * tair
    # Latent heat of vaporization (J/kg)
    Lv = 1000.0 * (2501 - 2.38 * tair)
    # Convert incoming radiation to J/m²
    rs_in_j = rs_in * 10_000.0
    # Reference evapotranspiration (mm/day)
    et_makkink = (1000.0 * 0.65 * d) / ((d + gamma) * 1000.0 * Lv) * rs_in_j

    return et_makkink



from string import Formatter


def has_placeholders(pattern: str) -> bool:
    """Return True als pattern één of meer {veld}-placeholders bevat."""
    for _, field_name, _, _ in Formatter().parse(pattern):
        if field_name:
            return True
    return False



import importlib
import pkgutil
from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version(__name__)
except PackageNotFoundError:
    __version__ = "0.0.0"  # for development


# import all modules dynamically
from . import config
from . import setup
from . import io
from . import postprocessing
from . import prepgrids
from . import submodels
from . import utils
from . import workflow
from .workflow.runner import FluxParkRunner

__all__ = [
    "config",
    "setup",
    "io",
    "postprocessing",
    "prepgrids",
    "utils",
    "workflow",
    "FluxParkRunner",
]

for loader, module_name, is_pkg in pkgutil.iter_modules(submodels.__path__):
    if module_name == "__init__":
        continue  # skip __init__.py

    full_module_name = f"{submodels.__name__}.{module_name}"
    module = importlib.import_module(full_module_name)

    for attr in dir(module):
        if not attr.startswith("_"):
            globals()[attr] = getattr(module, attr)
            __all__.append(attr)



from .interpolation import interpolate_rain, interpolate_makkink
from .prep_grids import load_fluxpark_raster_inputs, apply_evaporation_parameters

__all__ = [
    "interpolate_rain",
    "interpolate_makkink",
    "load_fluxpark_raster_inputs",
    "apply_evaporation_parameters",
]



"""
Computation of soil evaporation.

This module provides functions for estimating soil evaporation based on rain or
throughfall and potential evaporation.

Functions
---------
soilevap_boestenstroosnijder(throughfall, epot, beta_m, sum_ep_old, sum_ea_old)
    Calculates actual soil evaporaiton.
"""

import numpy as np


def soilevap_boestenstroosnijder(throughfall, epot, beta_m, sum_ep_old,
                                 sum_ea_old):
    """
    Calculate actual soil evaporation.

    Based on the method published by Boesten and Stroosnijder (1986).
    Supports scalar, 1D, and 2D array inputs but at least all in the same shape

    Parameters
    ----------
    throughfall : float or ndarray
        Throughfall [mm].
    epot : float or ndarray
        Potential soil evaporation [mm].
    beta_m : float or ndarray
        Boesten parameter in [m^0.5]; e.g., 0.5 in SWAP.
    sum_ep_old : float or ndarray
        Previous cumulative potential evaporation [mm].
    sum_ea_old : float or ndarray
        Previous cumulative actual evaporation [mm].

    Returns
    -------
    ea : float or ndarray
        Actual soil evaporation [mm].
    sum_ep : float or ndarray
        Updated cumulative potential evaporation [mm].
    sum_ea : float or ndarray
        Updated cumulative actual evaporation [mm].
    """
    # Ensure all inputs are arrays (at least 1D)
    tf = np.atleast_1d(throughfall)
    ep = np.atleast_1d(epot)
    beta = np.atleast_1d(beta_m)
    sum_ep_old = np.atleast_1d(sum_ep_old)
    sum_ea_old = np.atleast_1d(sum_ea_old)

    # Convert beta to mm^0.5
    beta_mm05 = np.sqrt((beta ** 2) * 1000.0)

    # Initialize arrays
    ea = np.zeros_like(tf, dtype="float32")
    sum_ep = sum_ep_old.copy()
    sum_ea = sum_ea_old.copy()

    # Determine condition for reset
    reset = (tf - ep) > sum_ep_old
    sum_ep[reset] = 0.0
    sum_ea[reset] = 0.0

    # Condition 1: no excess rain
    cond1 = tf < ep
    delta = ep - tf
    sum_ep[cond1] += delta[cond1]
    evap_limit = beta_mm05 * np.sqrt(sum_ep)
    over_limit = cond1 & (sum_ep > evap_limit)
    sum_ea[cond1] = sum_ep[cond1]
    sum_ea[over_limit] = evap_limit[over_limit]
    ea[cond1] = tf[cond1] + sum_ea[cond1] - sum_ea_old[cond1]
    ea[ea < 0.0] = 0.0

    # Update cumulative values for cond1
    sum_ep_old[cond1] = sum_ep[cond1]
    sum_ea_old[cond1] = sum_ea[cond1]

    # Condition 2: excess rain
    cond2 = ~cond1
    ea[cond2] = ep[cond2]
    sum_ea[cond2] = sum_ea_old[cond2] - (tf[cond2] - ea[cond2])
    sum_ea[cond2] = np.maximum(sum_ea[cond2], 0.0)
    sum_ep_calc = (sum_ea[cond2] ** 2) / (beta_mm05[cond2] ** 2)
    sum_ep[cond2] = np.maximum(sum_ep_calc, sum_ea[cond2])

    # Update cumulative values for cond2
    sum_ep_old[cond2] = sum_ep[cond2]
    sum_ea_old[cond2] = sum_ea[cond2]

    # Return same shape as input
    if throughfall.ndim == 1:
        return ea.flatten(), sum_ep.flatten(), sum_ea.flatten()
    else:
        return ea, sum_ep, sum_ea



import fluxpark as flp
import numpy as np
import logging
from concurrent.futures import ThreadPoolExecutor
import os
from osgeo import gdal


def write_output_tif(
    variable,
    filename,
    landuse_map,
    outdir,
    x_min,
    y_max,
    cellsize,
    epsg_code,
    dtype=gdal.GDT_Float32,
):
    """
    Write a single variable array as GeoTIFF, applying masks and scaling.

    Parameters
    ----------
    variable : np.ndarray
        2D array to write.
    filename : str
        Output filename (e.g. '20250601-rain.tif').
    landuse_map : np.ndarray
        2D landuse class map for masking sea/zero.
    outdir : str or Path
        Directory to write output file.
    x_min : float
        Upper-left X coordinate (corner of top-left pixel).
    y_max : float
        Upper-left Y coordinate (corner of top-left pixel).
    cellsize : float
        Pixel resolution (assumed square).
    epsg_code : int
        EPSG code for projection (e.g. 28992).
    dtype : int, optional
        GDAL data type constant (default: GDT_Float32).
    """
    # convert soil_cov fraction to percent
    if "soil_cov" in filename:
        variable = variable * 100.0

    # mask sea and zero landuse
    mask = (landuse_map == 17) | (landuse_map == 0)
    variable[mask] = np.nan

    # replace NaNs with nodata flag
    variable = np.where(np.isnan(variable), -9999, variable)

    # round to 3 decimals
    variable = np.around(variable, 3)

    flp.io.write_geotiff(
        outdir,
        filename,
        variable.astype(np.float32),
        x_min,
        y_max,
        cellsize,
        epsg_code,
        nodata_value=-9999,
        dtype=dtype,
    )


def write_all_tiffs(
    date,
    out_par_list,
    conv_output,
    daily_output,
    cum_output,
    landuse_map,
    outdir,
    x_min,
    y_max,
    cellsize,
    epsg_code,
    only_yearly_output=False,
    parallel=False,
    max_workers=None,
):
    """
    Write all requested parameters as GeoTIFFs for a given date.

    Parameters
    ----------
    date : datetime
        Current simulation date.
    out_par_list : list of str
        FluxPark parameter names to write.
    conv_output : dict
        Mapping from fluxpark param to variable name.
    daily_output : dict
        Daily variable arrays.
    cum_output : dict
        Cumulative variable values.
    landuse_map : np.ndarray
        Land use mask array.
    outdir : str or Path
        Directory for output files.
    x_min, y_max : float
        Upper-left corner coordinates.
    cellsize : float
        Resolution of output grid.
    epsg_code : int
        EPSG projection code.
    only_yearly_output : bool, optional
        Skip writing except year-end (default: False).
    parallel : bool, optional
        Use threaded output (default: False).
    max_workers : int, optional
        the amount of parallel workers, by default it will be derived from your machine
    """
    date_str = date.strftime("%Y%m%d")
    dtype = gdal.GDT_Float32

    # skip until year end if requested
    if only_yearly_output:
        if not (date.month == 12 and date.day == 31):
            return

    # prepare tasks
    tasks = []
    for par in out_par_list:
        filename = f"{date_str}-{par}.tif"
        var = conv_output[par]
        if var not in daily_output.keys() and var not in cum_output.keys():
            logging.warning(f"{var} not in the output dicts")
        array = cum_output[var] if var.endswith("_c") else daily_output[var]
        tasks.append((array, filename))

    # wrapper
    def _worker(args):
        arr, fname = args
        write_output_tif(
            arr,
            fname,
            landuse_map,
            outdir,
            x_min,
            y_max,
            cellsize,
            epsg_code,
            dtype,
        )

    if parallel:
        if not max_workers:
            max_workers = min((os.cpu_count() or 4), 64)
        with ThreadPoolExecutor(max_workers=max_workers) as exec:
            list(exec.map(_worker, tasks))
    else:
        for t in tasks:
            _worker(t)
